#include <iostream>
#include <stdlib.h>

using namespace std;

// Function to perform LU factorization
// U - Upper triangular matrix (modified in-place)
// L - Lower triangular matrix
// size - Size of the square matrix
void LU_factorization(float **U, float **L, int size) {
    for (int k = 0; k < size; ++k) {
        // Check for singularity
        if (U[k][k] == 0) {
            throw runtime_error("Matrix is singular!");
        }
        // Perform row operations to zero out elements below the diagonal
        for (int i = k + 1; i < size; i++) {
            double factor = U[i][k] / U[k][k];
            L[i][k] = factor; // Store the factor in the L matrix

            for (int j = k; j < size; j++) {
                U[i][j] -= factor * U[k][j];
            }
        }
    }
}

int main(int argc, const char *argv[]) {
    // Check for correct number of command-line arguments
    if (argc != 3) {
        fprintf(stderr, "Please provide the correct number of arguments.");
        exit(-1);
    }
    string out_directory = "../out/"; // Output directory
    FILE *f = fopen(argv[1], "r");    // Open the input file

    int size_matrix;
    fscanf(f, "%d", &size_matrix); // Read the matrix size

    // Allocate memory for matrices A, L, U, and P
    float **U, **L, **A;
    A = (float **)malloc(size_matrix * sizeof(*A));
    L = (float **)malloc(size_matrix * sizeof(*L));
    U = (float **)malloc(size_matrix * sizeof(*U));

    // Check for memory allocation failure
    if (!A || !L || !U) {
        fprintf(stderr, "Malloc failed");
        exit(-1);
    }

    // Allocate memory for each row of the matrices
    for (int i = 0; i < size_matrix; ++i) {
        A[i] = (float *)malloc(size_matrix * sizeof(**A));
        L[i] = (float *)malloc(size_matrix * sizeof(**L));
        U[i] = (float *)malloc(size_matrix * sizeof(**U));
    }

    // Initialize matrices and read input matrix A from file
    for (int i = 0; i < size_matrix; ++i) {
        for (int j = 0; j < size_matrix; ++j) {
            int aux;
            fscanf(f, "%d", &(aux));
            A[i][j] = (float)aux; // Initialize A
            U[i][j] = (float)aux; // Initialize U as a copy of A
        }
        L[i][i] = 1.0;
    }

    fclose(f); // Close the input file

    // Perform LU factorization
    LU_factorization(U, L, size_matrix);

    // Prepare output file
    out_directory.append("serial_");
    out_directory.append(argv[2]);
    f = fopen(out_directory.c_str(), "w");

    // Write lower triangular matrix L to the output file
    fprintf(f, "L:\n");
    for (int i = 0; i < size_matrix; ++i) {
        for (int j = 0; j < size_matrix; ++j) {
            fprintf(f, "%f ", L[i][j]);
        }
        fprintf(f, "\n");
    }
    fprintf(f, "\n");

    // Write upper triangular matrix U to the output file
    fprintf(f, "U:\n");
    for (int i = 0; i < size_matrix; ++i) {
        for (int j = 0; j < size_matrix; ++j) {
            fprintf(f, "%f ", U[i][j]);
        }
        fprintf(f, "\n");
    }

    // Free allocated memory
    for (int i = 0; i < size_matrix; ++i) {
        free(A[i]);
        free(L[i]);
        free(U[i]);
    }
    free(A);
    free(L);
    free(U);

    fclose(f); // Close the output file
    return 0;
}