## Ilinca Sebastian Ionut - 341C1
## Negru Mihai - 343C1

# LU Factorization

## Important

The code was tested on Linux Mint 22.

The makefile runs 2 tests that can be found in the input directory.
The output directory is created automatically from the makefile.

Requirements: `apt install openmpi-bin libopenmpi-dev g++`

Commands:
* make
* make run

## MPI LU Factorization Program

This program implements LU factorization using the Message Passing Interface (MPI) to take advantage of parallel computation across multiple processes.

### Code Overview

#### Key Components

1. **LU Factorization Function (`LU_factorization`)**
   - Broadcasts the current pivot row from the master process to all other processes.
   - Computes the L matrix's factors and updates the U matrix's rows below the pivot.
   - Synchronizes all processes using MPI_Barrier.

2. **Main Function (`main`)**
   - Master Process (Rank 0):
      - Reads the matrix from the input file.
      - Initializes the L and U matrices.
      - Broadcasts the matrix size to all processes.
      - Calls LU_factorization to compute the LU decomposition.
      - Writes the L and U matrices to an output file in the out/ directory.

   - Worker Processes:
      - Participate in the computation by receiving broadcasted data and performing matrix updates.
      - Synchronize with the master process.

#### Input and Output

- **Input**: The input file specifies the matrix size and elements in row-major order.
- **Output**: The matrices L, and U are written to an output file with a filename prefixed by "mpi_" and the specified suffix.

## OpenMP LU Factorization Program

This program implements LU factorization using OpenMP to exploit parallelism on multi-core processors.

### Code Overview

#### Key Components

1. **LU Factorization Function (`LU_factorization`)**
   - Performs the core LU decomposition.
   - Uses OpenMP to parallelize the elimination of elements below the diagonal in each column.
   - Ensures each thread independently handles its assigned rows to avoid conflicts.

2. **Main function**
   - Sets the number of threads for OpenMP parallelization.

#### Input and Output

- **Input**: The input file specifies the size of the matrix and its elements in row-major order.
- **Output**: The matrices L and U are written to an output file with a filename prefixed by "openmp_" and the provided suffix

## Hybrid MPI and OpenMP LU Factorization Program

This program implements LU factorization using a hybrid approach that combines MPI for distributed computation across multiple
nodes and OpenMP for parallel computation within each node.

### Code Overview

#### Key Components

1. **LU Factorization Function (`LU_factorization`)**
   - Iterates through the matrix column by column to compute the decomposition.
   - Uses OpenMP to parallelize row updates within each process.
   - Ensures synchronization and data sharing using MPI across processes.

3. **Main Function (`main`)**
   - Master Process (Rank 0):
      - Reads the matrix from the input file.
      - Initializes the L and U matrices.
      - Broadcasts the matrix size to all processes.
      - Calls LU_factorization to compute the LU decomposition.
      - Writes the L and U matrices to an output file in the out/ directory.

   - Worker Processes:
      - Receive the matrix size from the master process.
      - Participate in computation for LU factorization using OpenMP.

#### Input and Output

- **Input**: The input file specifies the size of the matrix and its elements in row-major order.
- **Output**: The matrices L and U are written to an output file with a filename prefixed by "hybrid_" and the provided suffix.

## LU Factorization with Pthreads

This program implements LU factorization using Pthreads to exploit parallelism across multiple threads.
### Code Overview

#### Key Components

1. **Worker Function (`workers`)**
   - Processes a subset of rows for the current step (`k`).
   - Each thread updates its assigned rows in parallel.
   - Computes the multiplier (`L[i][k]`) and updates the upper triangular matrix (`U`).

2. **LU Factorization Function (`LU_pthreads`)**
   - Orchestrates the LU decomposition process.
   - Divides rows among threads and creates threads for parallel computation.
   - Ensures synchronization by joining all threads before moving to the next step.

3. **Main Function (`main`)**
   - Handles input/output operations.
   - Reads the input matrix from a file and initializes matrices (L and U).
   - Invokes the LU factorization function and measures execution time.
   - Writes the results (L and U matrices) to an output file.
   - Manages memory allocation and deallocation to prevent leaks.